"""
扫描evals中的client，并返回一个client实例
扫描自身client，并返回一个client实例

俩个客户端，尽量结构一致
"""
import importlib.util
import inspect
import os
import logging
import sys
from pathlib import Path

from dacite import from_dict
from llmbench.inference.clients.request_data import Request

logger = logging.getLogger(__name__)


def import_module_by_file(file_path: str, module_name: str = ""):
    """Import a module from a file path."""
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    if module_name is not None:
        sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


class GeneralClient:

    def __init__(self, model, extra_params):
        self.model = model
        self.extra_params = extra_params or {}
        self.client_path = os.environ.get("client_path") or self.extra_params.get("client_path") # 测试用extra_params, 真实环境用env文件
        self.client = None
        self.is_self_client = True
        self.get_client()

    def get_client(self):
        client = self._get_self_client(self.model)
        if client is None:
            self.is_self_client = False
            if os.path.exists(self.model):
                if "llmbench_client" in self.model:
                    self.is_self_client = True
                client = self._get_file_client(self.model)
            else:
                client = self._get_eval_client(self.model)
        if client is None:
            raise ValueError(f"Could not find client for model: {self.model} or file_path:{self.model} or Model_Name is {self.model}")
        logger.info(f"current model is :{client}, is llmbench model: {self.is_self_client}")
        if self.is_self_client:
            self.client = client(**self.extra_params)
        else:
            self.client = client(model_extra_options=self.extra_params)

    def find_client_class(self, super_classes):
        super_class_names = [i.__name__ for i in super_classes]
        if "Client" in super_class_names and "ABC" in super_class_names:
            return True
        return False

    def _get_file_object(self, client_file_path: str, module_name: str):
        module = import_module_by_file(client_file_path, module_name)
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and obj.__name__ != "Client" and self.find_client_class(inspect.getmro(obj)):
                return obj
        return None

    def _get_file_client(self, client_file_path: str):
        client_dir, module_name = os.path.split(client_file_path)
        client = self._get_file_object(client_file_path, module_name)  # 文件形式加载
        if client is None:
            logger.warning(f"Could not find client for model: {client_file_path} "
                           f"file_path: {client_file_path}, try to load by driver")
        else:
            client = client
        return client

    def _get_self_client(self, model: str):
        load_client_info, self.is_self_client = LoadClient(self.client_path).get_file_client()
        if model in load_client_info:
            client = load_client_info.get(model)
            return client
        else:
            logger.warning(f"Could not find client for model in llmbench: {model}")

    def _get_eval_client(self, model):
        try:
            from evals.clients.client_auto import Model_Client
        except ImportError:
            logger.warning(f"Could not find client for model in evals: {model}")
            return
        if "/" in model:
            model_name, engine = model.split("/")
            if model_name in Model_Client:
                for _client in Model_Client[model_name]:
                    if engine in _client.get("engine", []):
                        client = _client.get("client")
                        return client

    def send_data(self, prompt, **kwargs):
        if self.is_self_client:
            result = self.client.make_request(from_dict(Request, {
                "model": self.model,
                "prompt": prompt,
                **kwargs
            }))
            result = result[0]
        else:
            result = self.client.make_request(from_dict(Request, {
                "model": self.model,
                "prompt": prompt,
                **kwargs
            }))
            if result:
                result = result.raw_response
        return result


class LoadClient:
    """
    根据模型自身定义的属性查找client,并返回映射关系
    """
    def __init__(self, client_file_path=None):
        self.client_file_path = client_file_path

    def _get_file_object(self, client_file_path: str, module_name: str):
        module = import_module_by_file(client_file_path, module_name)
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and hasattr(obj, "Model_Name"):
                return {obj.Model_Name: obj}

    def get_file_client(self):
        is_self_client = True
        if not self.client_file_path:
            self.client_file_path = os.path.join(os.path.dirname(__file__), "clients")
        else:
            if self.client_file_path.endswith("eval_client"):
                is_self_client = False

        client_mapping = {}
        for client_file in os.listdir(self.client_file_path):
            tmp_client = Path(client_file)
            if tmp_client.stem.startswith("client") and tmp_client.suffix == ".py":
                map_info = self._get_file_object(os.path.join(self.client_file_path, client_file), client_file)
                if map_info:
                    client_mapping.update(**map_info)
        if not client_mapping:
            logger.warning(f"Could not find client by Model_Name.")
        return client_mapping, is_self_client



import json

from gevent import monkey

monkey.patch_all()
import gevent
from gevent import queue, pool
import traceback
import time
from typing import Optional, Dict, Any, List
from tqdm import tqdm
import logging
from llmbench.data import get_file_lines
from llmbench.report import Report, HtmlReport
from llmbench.inference.general_client import GeneralClient

logger = logging.getLogger(__name__)


class TokenBenchmark(object):

    def __init__(
            self,
            dataset: str,
            num_concurrent_requests: int,
            test_timeout: int,
            model: str,  # 模型名称或模型路径
            max_request_sample_num: int = -1,  # 最大请求样本数，默认-1, 无限制
            result_dir: str = None,  # 结果保存路径
            need_calc_token: bool = False,
            extra_params: Optional[Dict[str, Any]] = None,
    ):
        self.pool = pool.Pool(num_concurrent_requests)
        self.dataset = dataset
        self.duration = test_timeout
        self.model = model
        self.stop_flag = False
        if extra_params is None:
            extra_params = {}
        if isinstance(extra_params, str) and extra_params.startswith("{"):
            extra_params = json.loads(extra_params)
        self.extra_params = extra_params
        self.max_request_sample_num = max_request_sample_num if max_request_sample_num != -1 else max_request_sample_num
        self.report = Report(model.replace("/", "_"), result_dir, **{"start_t": time.time()})
        self.num_concurrent_requests = num_concurrent_requests
        self.max_queue_size = num_concurrent_requests * 10
        self.job_queue = queue.Queue(maxsize=self.max_queue_size)
        self.report_queue = queue.Queue()
        self.client = GeneralClient(self.model, self.extra_params)
        self.need_calc_token = need_calc_token
        self.last_report_list = []  # 记录最后一次计算结果

    def get_prompt_data(self, max_line_num: int = 5000):
        """获取测试数据"""
        data_lines = get_file_lines(self.dataset, max_line_num)
        return data_lines

    def send_data(self, prompt, **kwargs):
        return self.client.send_data(prompt, **kwargs)

    def clac_report(self):
        """计算报告"""
        start_time = time.time()
        report_list = []
        while time.time() - start_time < self.duration and not self.stop_flag:
            try:
                if self.report_queue.empty():
                    gevent.sleep()
                    continue
                report = self.report_queue.get(block=False)
                if report:
                    report_list.append(report)
                if len(report_list) > 20:
                    self.report.save_result(report_list)
                    self.last_report_list = [report_list[-1]]
                    report_list.clear()
            except Exception as e:
                logger.error(traceback.format_exc())
                gevent.sleep()
        self.report.save_result(self.last_report_list, last=True)

    def worker(self):
        """工作线程"""
        start_time = time.time()
        while time.time() - start_time < self.duration and not self.stop_flag:
            try:
                if self.job_queue.empty():
                    gevent.sleep()
                    continue
                task = self.job_queue.get(timeout=1)
                result = self.send_data(task)
                if result:
                    self.report_queue.put_nowait(result)
            except (Exception, queue.Empty) as e:
                logger.error(traceback.format_exc())
                gevent.sleep()

    def producer(self):
        """生产者线程"""
        start_time = time.time()
        iter_num = 0
        prompt = self.get_prompt_data(self.max_request_sample_num)
        max_line_num = self.max_request_sample_num if len(prompt) >= self.max_request_sample_num else len(prompt)
        while time.time() - start_time < self.duration and not self.stop_flag:
            try:
                if not self.job_queue.full():
                    iter_num += 1
                    self.job_queue.put_nowait(prompt[iter_num % max_line_num])
                else:
                    gevent.sleep()  # 等待队列有空间
            except queue.Full:
                gevent.sleep()  # 等待队列有空间

    def wait_all_task(self, tasks: List[gevent.Greenlet]):
        """等待所有任务完成"""
        gevent.joinall(tasks)

    def start_all_task(self) -> List[gevent.Greenlet]:
        job_events = []
        try:
            # 生产线程
            job_events.append(gevent.spawn(self.producer))
            # 报告线程
            job_events.append(gevent.spawn(self.clac_report))
            # 工作线程
            job_events.extend(
                [gevent.spawn(self.worker) for _ in range(self.num_concurrent_requests)])
            # 工作线程  报错，后续看看
            # self.pool.spawn(self.worker, range(self.num_concurrent_requests))
        except Exception as e:
            print(e)
        return job_events

    def start_process_monitor(self):
        """进度监控"""

        def update_process(pbar, process):
            pbar.n = int(process)
            pbar.refresh()

        start_time = time.time()
        with tqdm(total=self.duration, desc="执行进度") as pbar:
            while time.time() - start_time < self.duration:
                gevent.sleep(5)
                update_process(pbar, time.time() - start_time)
            update_process(pbar, time.time() - start_time)

    def run_token_benchmark(self):
        # start_time = time.time()
        tasks = self.start_all_task()
        # 启动进度监听
        self.start_process_monitor()
        # 等待所用任务完成
        self.wait_all_task(tasks)
        # all_cost_time = time.time() - start_time
        # print("执行总耗时：{}".format(all_cost_time))
        time.sleep(1)  # 等待数据更新
        HtmlReport(self.report).make_report()




import json
import random
import time
import requests
from llmbench.inference.clients.client_abc import Client
from llmbench.common.constants import METRICS
from llmbench.utils import generate_token


class TestAPIClient(Client):
    """Client for OpenAI Chat Completions API."""
    Model_Name = "test"

    def __init__(self, **kwargs):
        pass

    def get_token(self, auth):
        auth = auth or {}
        apikey = auth.get("API_KEY", "65ae213bf7b40c25fbe169bb.JxNn87L/KgLO7rWGOTHOnTzW0NCnJQaK")
        return generate_token(apikey, 3600)

    def make_request(self, request_config):
        prompt = request_config.prompt
        prompt_len = random.randint(len(prompt), 100)
        jwt_token = self.get_token(request_config.auth)
        address = request_config.url or "http://127.0.0.1:5000/stream"

        body = {
            "modelId": request_config.model or "jiutian-lan",
            "prompt": prompt,
            "params": {"temperature": 0.8, "top_p": 0.95},
            "history": [],
            "stream": True
        }
        headers = {'content-type': "application/json", "Authorization": "Bearer " + jwt_token}

        time_to_next_token = []
        ttft = 0
        tokens_received = random.randint(1, 10)
        error_response_code = -1
        error_msg = ""
        generated_text = ""
        output_throughput = 0
        total_request_time = 0
        start_time = time.monotonic()
        most_recent_received_token_time = time.monotonic()

        metrics = {METRICS.ERROR_CODE: None, METRICS.ERROR_MSG: "",
                   METRICS.START_TIME: time.time(), METRICS.START_TIME_M: start_time}

        try:
            with requests.get(
                    address,
                    json=body,
                    stream=True,
                    timeout=180,
                    headers=headers,
            ) as response:
                time.sleep(random.randint(1, 3))
                if response.status_code != 200:
                    error_msg = response.text
                    error_response_code = response.status_code
                    response.raise_for_status()
                for chunk in response.iter_lines(chunk_size=None):
                    if not chunk:
                        continue

                    chunk = chunk.decode("utf-8").strip()
                    stem = "data:"
                    if not chunk.startswith(stem):
                        continue
                    chunk = chunk.split(stem)[-1]
                    try:
                        data = json.loads(chunk)
                    except Exception as e:
                        continue

                    tokens_received += 1

                    if "delta" not in data.keys():
                        error_msg = data["message"]
                        error_response_code = data["code"]
                        raise RuntimeError(data["message"])

                    delta = data["delta"]
                    if delta:
                        if not ttft:
                            time.sleep(random.random())
                            ttft = time.monotonic() - start_time
                            time_to_next_token.append(ttft)
                        else:
                            time_to_next_token.append(
                                time.monotonic() - most_recent_received_token_time
                            )
                        most_recent_received_token_time = time.monotonic()
                        generated_text += delta

            total_request_time = time.monotonic() - start_time
            output_throughput = tokens_received / total_request_time

        except Exception as e:
            metrics[METRICS.ERROR_MSG] = error_msg
            metrics[METRICS.ERROR_CODE] = error_response_code
            print(f"Warning Or Error: {e}")
            print(error_response_code)

        metrics[METRICS.END_TIME_M] = time.monotonic()
        metrics[METRICS.END_TIME] = time.time()
        # This should be same as METRICS[METRICS.E2E_LAT]. Leave it here for now
        metrics[METRICS.INTER_TOKEN_LAT] = sum(time_to_next_token)
        metrics[METRICS.TTFT] = ttft
        metrics[METRICS.E2E_LAT] = total_request_time
        metrics[METRICS.REQ_OUTPUT_THROUGHPUT] = output_throughput
        metrics[METRICS.NUM_TOTAL_TOKENS] = tokens_received + prompt_len
        metrics[METRICS.NUM_OUTPUT_TOKENS] = tokens_received
        metrics[METRICS.NUM_INPUT_TOKENS] = prompt_len
        return metrics, generated_text, request_config



import json
import os
import time
import requests
from llmbench.common.constants import METRICS
from llmbench.inference.clients.client_abc import Client


class OpenAIChatCompletionsClient(Client):
    """Client for OpenAI Chat Completions API."""
    Model_Name = "openai"

    def __init__(self, **kwargs):
        pass

    def make_request(self, request_config):
        prompt = request_config.prompt
        prompt, prompt_len = prompt

        message = [
            {"role": "system", "content": ""},
            {"role": "user", "content": prompt},
        ]
        model = request_config.model
        body = {
            "model": model,
            "messages": message,
            "stream": True,
        }
        sampling_params = request_config.sampling_params
        body.update(sampling_params or {})
        time_to_next_token = []
        tokens_received = 0
        ttft = 0
        error_response_code = -1
        generated_text = ""
        error_msg = ""
        output_throughput = 0
        total_request_time = 0

        metrics = {METRICS.ERROR_CODE: None, METRICS.ERROR_MSG: ""}

        start_time = time.monotonic()
        most_recent_received_token_time = time.monotonic()
        address = os.environ.get("OPENAI_API_BASE")
        if not address:
            raise ValueError("the environment variable OPENAI_API_BASE must be set.")
        key = os.environ.get("OPENAI_API_KEY")
        if not key:
            raise ValueError("the environment variable OPENAI_API_KEY must be set.")
        headers = {"Authorization": f"Bearer {key}"}
        if not address:
            raise ValueError("No host provided.")
        if not address.endswith("/"):
            address = address + "/"
        address += "chat/completions"
        try:
            with requests.post(
                address,
                json=body,
                stream=True,
                timeout=180,
                headers=headers,
            ) as response:
                if response.status_code != 200:
                    error_msg = response.text
                    error_response_code = response.status_code
                    response.raise_for_status()
                for chunk in response.iter_lines(chunk_size=None):
                    chunk = chunk.strip()

                    if not chunk:
                        continue
                    stem = "data: "
                    chunk = chunk[len(stem) :]
                    if chunk == b"[DONE]":
                        continue
                    tokens_received += 1
                    data = json.loads(chunk)

                    if "error" in data:
                        error_msg = data["error"]["message"]
                        error_response_code = data["error"]["code"]
                        raise RuntimeError(data["error"]["message"])
                        
                    delta = data["choices"][0]["delta"]
                    if delta.get("content", None):
                        if not ttft:
                            ttft = time.monotonic() - start_time
                            time_to_next_token.append(ttft)
                        else:
                            time_to_next_token.append(
                                time.monotonic() - most_recent_received_token_time
                            )
                        most_recent_received_token_time = time.monotonic()
                        generated_text += delta["content"]

            total_request_time = time.monotonic() - start_time
            output_throughput = tokens_received / total_request_time

        except Exception as e:
            metrics[METRICS.ERROR_MSG] = error_msg
            metrics[METRICS.ERROR_CODE] = error_response_code
            print(f"Warning Or Error: {e}")
            print(error_response_code)

        metrics[METRICS.INTER_TOKEN_LAT] = sum(time_to_next_token) #This should be same as metrics[metrics.E2E_LAT]. Leave it here for now
        metrics[METRICS.TTFT] = ttft
        metrics[METRICS.E2E_LAT] = total_request_time
        metrics[METRICS.REQ_OUTPUT_THROUGHPUT] = output_throughput
        metrics[METRICS.NUM_TOTAL_TOKENS] = tokens_received + prompt_len
        metrics[METRICS.NUM_OUTPUT_TOKENS] = tokens_received
        metrics[METRICS.NUM_INPUT_TOKENS] = prompt_len

        return metrics, generated_text, request_config



from pathlib import Path

COMMON_MODEL_PATH = Path.home() / ".cache" / "huggingface" / "hub"
TOKENIZER_MODEL_PATH = COMMON_MODEL_PATH / "models--hf-internal-testing--llama-tokenizer"


class TOKENIZER_MODEL:
    """# 0: openai tiktoken 1: models--hf-internal-testing--llama-tokenizer"""
    tiktoken = 0
    llama = 1


default_tokenizer_model = TOKENIZER_MODEL.tiktoken


class METRICS:
    START_TIME = "start_time"
    END_TIME = "end_time"
    START_TIME_M = "start_time_m"
    END_TIME_M = "end_time_m"
    INTER_TOKEN_LAT = "inter_token_latency_s"
    TTFT = "ttft_s"
    E2E_LAT = "end_to_end_latency_s"
    NUM_INPUT_TOKENS = "number_input_tokens"
    NUM_OUTPUT_TOKENS = "number_output_tokens"
    NUM_TOTAL_TOKENS = "number_total_tokens"
    REQ_OUTPUT_THROUGHPUT = "request_output_throughput_token_per_s"
    ERROR_MSG = "error_msg"
    ERROR_CODE = "error_code"
    ERROR_CODE_FREQ = "error_code_frequency"
    NUM_ERRORS = "number_errors"
    OUTPUT_THROUGHPUT = "mean_output_throughput_token_per_s"
    NUM_COMPLETED_REQUESTS = "num_completed_requests"
    COMPLETED_REQUESTS_PER_MIN = "num_completed_requests_per_min"
    ERROR_RATE = "error_rate"
    NUM_REQ_STARTED = "num_requests_started"
    TOTAL_COST_TIME = "total_cost_time_s"
